/************************************************************************

    JSWED 2.00 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "guardedit.hxx"
#include "j128gedit.hxx"
#include "j128guard.hxx"
#include "jswdirbox.hxx"
#include "hexedit.hxx"
#include "vspritelist.hxx"

Jsw128GuardEditor::Jsw128GuardEditor(JswGame *g,jswByte *guard,VideoBitmap *bg, int x) 
	 : GuardianEditor(g, guard, bg, x)
{
	setupControls();

        m_labelAnim2 = new VideoLabel(m_x + 27 * CHAR_W, m_y + 22 * CHAR_H, "Fast");
        m_cbAnim2    = new VideoCheckbox(m_x + 32 * CHAR_W, m_y + 22 * CHAR_H);

        m_labelBounds = new VideoLabel(m_x + 8 * CHAR_W, m_y + 24 * CHAR_H, "Bounds");
        m_spinBounds[0] = new VideoSpinControl(m_x + 19 * CHAR_W, m_y + 24 * CHAR_H,
                                                7 * CHAR_W, CHAR_H + 4);
        m_spinBounds[1] = new VideoSpinControl(m_x + 27 * CHAR_W-4, m_y + 24 * CHAR_H,
                                                7 * CHAR_W, CHAR_H + 4);
        m_spinBounds[0]->setListener(this);
        m_spinBounds[1]->setListener(this);	
	m_cbAnim2->setListener(this);
}

Jsw128GuardEditor::~Jsw128GuardEditor()
{
	delete m_spinBounds[0];
	delete m_spinBounds[1];
	delete m_labelBounds;
	delete m_labelAnim2;
	delete m_cbAnim2;
	deleteControls();
}


void Jsw128GuardEditor::subEditChange(VideoControl *c)
{	
	if (c == m_spinBounds[0] || c == m_spinBounds[1])
	{
		((Jsw128Guard *)m_guard)->setBounds(m_gdata[6], m_gdata[7]);
	}
        int type = m_gdata[0] & 0x0F;

        switch(type)
        {
                case 1: case 2:  case  5: case  6:
                case 9: case 10: case 13: case 14:
                if (c == m_spinVstep)
                {
                        m_guard->setStep(m_gdata[4]);
                }
                if (c == m_spinY)
                {
                        m_guard->setY(m_gdata[3]);
                }
                if (c == m_spinAnim)
                {
                        m_guard->setAnim(m_spinAnim->getValue());
                        m_gdata[1] &= 0x1F;
                        m_gdata[1] |= ((m_spinAnim->getValue() & 0x07) << 5);
                }
                break;

                case 3:
                if (c == m_spinRopeLen)
                {
                        m_guard->setRopeLen(m_spinRopeLen->getValue());
                        m_gdata[4] = m_spinRopeLen->getValue();
                }
                if (c == m_spinRopeSwing)
                {
                        m_guard->setRopeSwing(m_spinRopeSwing->getValue());
                        m_gdata[7] = m_spinRopeSwing->getValue();
                }
                if (c == m_spinRopeX)
                {
                        int rx = m_spinRopeX->getValue();
                        if (rx < 0) m_gdata[1] = 0x80 | (-rx);
                        else        m_gdata[1] = rx;
                }
                break;
        }
}

int  Jsw128GuardEditor::subCheckboxSelect(VideoCheckbox *c, int value)
{
        int type = m_gdata[0] & 0x0F;

	if (c == m_cbAnim2 && (type ==2 || type == 10))
	{
		m_gdata[0] &= ~0x10;
		m_gdata[0] |= (value ? 0x10 : 0);
		((Jsw128Guard *)m_guard)->setFastAnim(value);
		return ZXE_OK;
	}
        if (c == m_cbDirection)
        {
                m_gdata[0] &= 0x7F;
                if (value) m_gdata[0] |= 0x80;
        }
	return ZXE_OK;
}

void Jsw128GuardEditor::changeType(void)
{
	int t = m_gdata[0] & 0x0F;
	
	removeVarChildren();
	switch(t)
	{
		// Horizontals and diagonals
		case 5: case 6: 
		case 13: case 14: 
			addChild(m_spinVstep);
                        m_spinVstep->bind(m_gdata + 4);
		case 1: case 9:
			addChild(m_bnSprite);
			addChild(m_labelY);
			addChild(m_spinY);
			m_spinY->bind(m_gdata + 3);
			addChild(m_chInk);
			addChild(m_cbBright);
			addChild(m_labelBright);
			addChild(m_labelAnim);
			addChild(m_spinAnim);
			m_cbDirection->setVertical(0);
			addChild(m_labelBounds);
			addChild(m_spinBounds[0]);
			addChild(m_spinBounds[1]);
			m_spinBounds[0]->bind(m_gdata + 6);
			m_spinBounds[1]->bind(m_gdata + 7);
			m_spinBounds[0]->setDelta(1);
			m_spinBounds[1]->setDelta(1);
			break;
		// Verticals
		case 2: case 10: 
			addChild(m_labelAnim2);
			addChild(m_cbAnim2);
		case 7: case 15:
			addChild(m_bnSprite);
			addChild(m_labelY);
			addChild(m_spinY);
			m_spinY->bind(m_gdata + 3);
			addChild(m_spinVstep);
			m_spinVstep->bind(m_gdata + 4);
			addChild(m_chInk);
			addChild(m_cbBright);
			addChild(m_labelBright);
                        addChild(m_labelAnim);
                        addChild(m_spinAnim); 
			m_cbDirection->setVertical(1);
			addChild(m_labelBounds);
                        addChild(m_spinBounds[0]);
                        addChild(m_spinBounds[1]);
                        m_spinBounds[0]->bind(m_gdata + 6);
                        m_spinBounds[1]->bind(m_gdata + 7);
                        m_spinBounds[0]->setDelta(2);
                        m_spinBounds[1]->setDelta(2);

			break;
		// Rope
		case 3: addChild(m_spinRopeX);
			addChild(m_labelRopeX);
			addChild(m_spinRopeLen);
			addChild(m_labelRopeLen);
			addChild(m_spinRopeSwing);
			addChild(m_labelRopeSwing);
			m_cbDirection->setVertical(0);
			break;
		// Arrow
		case 4: addChild(m_bmEdit);
			m_bmEdit->setBitmap(m_gdata + 6);
			addChild(m_spinStartPos);
			addChild(m_labelStartPos);
			m_spinStartPos->bind(m_gdata + 4);	
			m_cbDirection->setVertical(0);
			break;
	}
	redraw();
}

void Jsw128GuardEditor::removeVarChildren(void)
{
        removeChild(m_labelBounds);
        removeChild(m_spinBounds[0]);
        removeChild(m_spinBounds[1]);
        removeChild(m_cbAnim2);
        removeChild(m_labelAnim2);
	GuardianEditor::removeVarChildren();
}

void Jsw128GuardEditor::drawTsData(void)
{
        int rx = m_gdata[1] & 0x7F;
        if (m_gdata[1] & 0x80) rx = -rx;

        m_cbDirection->setCheck(m_gdata[0] & 0x80);
	switch(m_gdata[0] & 15)
	{
	        case 2: case 10:
	        m_cbAnim2->setCheck((m_gdata[0] & 0x10) ? 1 : 0);
	        case 1: case 9:
	        case 5: case 6: case 13: case 14:
	        m_spinAnim->setValue(m_gdata[1] >> 5);
	        break;

	        case 3:
	        m_spinRopeX->setValue(rx);
	        m_spinRopeLen->setValue(m_gdata[4]);
	        m_spinRopeSwing->setValue(m_gdata[7]);
	        break;
	}
}


int Jsw128GuardEditor::setGuardianType(void)
{
	static jswByte stdRope[] = {0x03, 0x22, 0x00, 0x00, 
				    0x20, 0x00, 0x83, 0x36 };
	static jswByte stdArrow[]= {0x04, 0x06, 0x00, 0x00,
				    0xd0, 0x00, 0x41, 0x00 };
	static int types[] = { 0, 1, 2, 3, 4, 5, 6, 7, 9, 13, 14 };
	VideoMenu vm48("Guardian type", "Blank", "Horizontal", "Vertical", 
			"Rope", "Arrow", NULL);
	VideoMenu vm128("Guardian type", 
			"Blank", 			// 0
			"Horizontal", 			// 1
			"Vertical",   			// 2
			"Rope",       			// 3
			"Arrow",      			// 4
			"Diagonal NW/SE", 		// 5
			"Diagonal NE/SW", 		// 6 
			"Vertical multicolour", 	// 7
			"Horizontal multicolour",	// 8
			"NW/SE multicolour", 		// 9
			"NE/SW multicolour");		// 10

	int rv, sel;

	preContextMenu();
	if (m_game->supportsExtGuards()) 
	{
		rv  = vm128.doModal();
		sel = vm128.getSelected();
	}
	else
	{
		rv  = vm48.doModal();
		sel = vm48.getSelected();
	}
	postContextMenu();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK) return ZXE_CONTINUE;
	
	int ot = m_gdata[0] & 0x0F;
	int nt = types[sel];
	switch(sel)
	{
		case 1:		
		case 5:
		case 6:
		case 8:
		case 9:
		case 10: if (ot == 1 || ot == 5  || ot == 6 || 
			    ot == 9 || ot == 13 || ot == 14)
			{
				m_gdata[0] &= 0xF0;
				m_gdata[0] |= nt;
			}
			else
			{
				m_gdata[0] = nt;
				m_gdata[6] = 19;
				m_gdata[7] = 29;
				m_gdata[4] = 0xFC;
			}
			break;	
		case 2:
		case 7:	if (ot == 2 || ot == 10 || ot == 7 || ot == 15)
			{
                                m_gdata[0] &= 0xF0;
                                m_gdata[0] |= nt;
			}
			else
			{
				m_gdata[0] = nt;
				m_gdata[6] = 0x80;
				m_gdata[7] = 0xE0;
				m_gdata[4] = 0x04;
			}
			break;
		case 3:	if (ot != 3) memcpy(m_gdata, stdRope, 8);
			break;
		case 4: if (ot != 4) memcpy(m_gdata, stdArrow, 8);
			break;
		default:
			m_gdata[0] &= 0xF0; m_gdata[0] |= nt;	
	}
	delete m_guard;
        m_guard = m_game->newGuardian(m_gdata, m_nx, 0);
	changeType();
	drawData(0);
	return ZXE_CONTINUE;
}

