/************************************************************************

    JSWED 2.00 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "j48roomdraw.hxx"



Jsw48RoomDraw::Jsw48RoomDraw(JswGame *game)
{
	m_game = game;
	m_ticker = 0;
	m_room = game->getStartRoom();
	m_ot   = NULL;
	m_haveTimer = 0;
	m_visible = 0;

	for (int n = 0; n < (8 + FG_MAX); n++) m_guards[n] = NULL;
	loadRoom(m_room);

        m_bmAir   = NULL;
        m_bmWater = NULL;
        m_bmEarth = NULL;
        m_bmFire  = NULL;
        m_bmRamp  = NULL;
        m_bmConvey= NULL;
	m_bmObject= NULL;
	m_bmRoom  = newVideoBitmap(512, 256);
	m_bmAnim  = newVideoBitmap(512, 256);

	m_drawnFixed = 0;

}

Jsw48RoomDraw::~Jsw48RoomDraw()
{
        if (m_bmAir)    delete m_bmAir;    m_bmAir    = NULL;
        if (m_bmWater)  delete m_bmWater;  m_bmWater  = NULL;
        if (m_bmEarth)  delete m_bmEarth;  m_bmEarth  = NULL;
        if (m_bmFire)   delete m_bmFire;   m_bmFire   = NULL;
        if (m_bmRamp)   delete m_bmRamp;   m_bmRamp   = NULL;
        if (m_bmConvey) delete m_bmConvey; m_bmConvey = NULL;
	if (m_bmObject) delete m_bmObject; m_bmObject = NULL;
	delete m_bmRoom;
	delete m_bmAnim;
	for (int n = 0; n < (8 + FG_MAX); n++) 
		if (m_guards[n]) delete m_guards[n]; 

        if (m_haveTimer) setTickListener(m_ot);
}




void Jsw48RoomDraw::drawFixed(void)
{
	m_drawnFixed = 1;
}

VideoBitmap *Jsw48RoomDraw::bitmapFromSprite(jswByte *sprite)
{
	int ink   = sprite[0] & 7;
	int paper = (sprite[0] >> 3) & 7;

	if (sprite[0] & 64)  { ink |= ZX_BRIGHT; paper |= ZX_BRIGHT; }
	return ::bitmapFromSprite(8, 8, sprite+1, 1, ink, paper);
}

#define CONVEYORANIM 4	/* 2 or even 1 look nicer, but 4 matches JSW best */

/* Overridables */
void Jsw48RoomDraw::drawCellType() { }
void Jsw48RoomDraw::drawCursor(void) { }
void Jsw48RoomDraw::drawGuardList(void) { }

void Jsw48RoomDraw::flashColours(void)
{
	++m_ticker;

	if (!m_visible) return;
	
	drawCellType();

	if ((m_ticker % 20)== 0)   drawData(1);

	if ( (m_ticker % 5)== 0  && m_bmConvey)
	{
		if (getRoom(m_room)[0xD6] == 1)
		{
                  m_bmConvey->getSurface()->rotateR(0, 0, 16, 2, CONVEYORANIM);
                  m_bmConvey->getSurface()->rotateL(0, 4, 16, 2, CONVEYORANIM);
		}
		else if (getRoom(m_room)[0xD6] == 0)
		{
		  m_bmConvey->getSurface()->rotateL(0, 0, 16, 2, CONVEYORANIM);
		  m_bmConvey->getSurface()->rotateR(0, 4, 16, 2, CONVEYORANIM);
		}
		drawConveyor();
		m_bmRoom->toSurface(m_bmAnim, 0, 0);
		for (int n = 0; n < (8 + FG_MAX); n++)
		{	
			if (m_guards[n])
			{
				m_guards[n]->move();
				m_guards[n]->draw(m_bmAnim);
			}
			else if (n < 8) n = 7; 
			else break;
		}
		drawObjects();
		drawCursor();
	        m_bmAnim->toScreen(m_x, m_y);
	}

}







void Jsw48RoomDraw::flashSprite(VideoBitmap *bmp, jswByte attr)
{
        int ink   = attr & 7;
        int paper = (attr >> 3) & 7;
	int im;

	if (!(attr & 128)) return; 

        if (attr & 64)  { ink |= ZX_BRIGHT; paper |= ZX_BRIGHT; }

	for (im = 1; im < 15; im++)
	{
		if (im != paper && im != ink && im != 8) break;
	}
	bmp->getSurface()->fillReplace(0, 0, 16, 16, paper, im);
	bmp->getSurface()->fillReplace(0, 0, 16, 16, ink, paper);
	bmp->getSurface()->fillReplace(0, 0, 16, 16, im, ink);
}

void Jsw48RoomDraw::drawData(int flashonly)
{
	jswByte *room = getRoom(m_room);

	if (!m_bmAir)    m_bmAir    = bitmapFromSprite(room + 0xA0);
	if (!m_bmWater)  m_bmWater  = bitmapFromSprite(room + 0xA9);
	if (!m_bmEarth)  m_bmEarth  = bitmapFromSprite(room + 0xB2);
	if (!m_bmFire)   m_bmFire   = bitmapFromSprite(room + 0xBB);
	if (!m_bmRamp)   m_bmRamp   = bitmapFromSprite(room + 0xC4);
	if (!m_bmConvey) m_bmConvey = bitmapFromSprite(room + 0xCD);
	if (!m_bmObject)
	{
		jswByte b = room[0xE0];
		room[0xE0] = 0x07;	/* White */
		m_bmObject = bitmapFromSprite(room + 0xE0);
		room[0xE0] = b;
	}

	if (flashonly)
	{
		flashSprite(m_bmAir,   room[0xA0]);
		flashSprite(m_bmWater, room[0xA9]);
		flashSprite(m_bmEarth, room[0xB2]);
		flashSprite(m_bmFire,  room[0xBB]);
		flashSprite(m_bmRamp,  room[0xC4]);
		flashSprite(m_bmConvey,room[0xCD]);
	}

	drawBasic(flashonly);
	drawRamp();
	if (!flashonly)
	{
		char sbuf[33];
		int n;

		drawConveyor();
		drawGuardList();	
		sprintf(sbuf, "%-32.32s", room + 0x80);
		for (n = 0; n < 32; n++)
		{
			if (sbuf[n] == ZXCH_COPY ) sbuf[n] = ZXK_COPY;
			if (sbuf[n] == ZXCH_POUND) sbuf[n] = ZXK_POUND;
		}
		videoScreen->drawText(m_x, m_y + 256, sbuf, ZX_BLACK,
			ZX_BRWHITE);
	        m_bmRoom->toSurface(m_bmAnim, 0, 0);
                for (n = 0; n < (8 + FG_MAX); n++)
                {
                        if (!m_guards[n]) break;
                        m_guards[n]->draw(m_bmAnim);
                }
		drawObjects();
		m_bmAnim->toScreen(m_x, m_y);
	}
}


void Jsw48RoomDraw::drawBasic(int flashonly)
{
        jswByte *room = getRoom(m_room);
        int x, y;

	for (y = 0; y < 16; y++) for (x = 0; x < 32; x++)
	{
		drawCell(room, x, y, flashonly);
	}
}




void Jsw48RoomDraw::drawCell(jswByte *room, int x, int y, int flashonly)
{
	jswByte cell = room[y * 8 + (x / 4)];

	cell = (cell << ((x % 4) * 2)) & 0xC0;

	switch(cell)
	{
		case 0x00: if (!flashonly || (room[0xA0] & 128))
			   	m_bmAir->toSurface(m_bmRoom, x*16, y*16);
			   break;
		case 0x40: if (!flashonly || (room[0xA9] & 128))
			   	m_bmWater->toSurface(m_bmRoom, x*16, y*16);
			   break;
		case 0x80: if (!flashonly || (room[0xB2] & 128))
                                m_bmEarth->toSurface(m_bmRoom, x*16, y*16);
			   break;
		case 0xC0: if (!flashonly || (room[0xBB] & 128))
                                m_bmFire->toSurface(m_bmRoom, x*16, y*16);
			   break;
	}
}



void Jsw48RoomDraw::drawRamp(int draw = 1)
{
	jswByte *room = getRoom(m_room);
        int len;

        int w=room[0xDB]+(256*room[0xDC]);
        int x = (w & 0x1F);
        int y = ((w >> 1) & 0xF0);

        if (room[0xDD] > 32) /* Damaged room */
                room[0xDD]=0;

        len = room[0xDD];
        if (room[0xDA] != 1) len = (-len);

        drawRamp(y,x,len,room, draw);
}


void Jsw48RoomDraw::drawRamp(int y, int x, int len, 
				jswByte *room, int draw)
{
        int n, step;

        if (len < 0) { len = (-len); step = (-1); } else step = 1;

        y = (y >> 4);

        for (n = 0; n < len; n++)
        {
                if (y >= 0 && y < 16 && x >= 0 && x < 32)
                {
                        if (draw)
				m_bmRamp->toSurface(m_bmRoom, 16 * x, 16 * y);
                        else	drawCell(room, x, y, 0);

                }
                x += step; --y;
	}
}

void Jsw48RoomDraw::drawConveyor(int draw)
{
        jswByte *room = getRoom(m_room);

	int len;
        int w=room[0xD7]+(256*room[0xD8]);
        int x = (w & 0x1F);
        int y = ((w >> 1) & 0xF0);

        if (room[0xD9] > 32) /* Damaged room */
                room[0xD9]=0;

	len = room[0xD9];

	drawConveyor(x, y, len, room, draw);
}

void Jsw48RoomDraw::drawConveyor(int x, int y, int len, jswByte *room, int draw)
{
	y = (y >> 4);

	for (int n = 0; n < len; n++)
	{
		if (y >= 0 && y < 16 && x >= 0 && x < 32)
		{
			if (draw) m_bmConvey->toSurface(m_bmRoom, x*16, y*16);
			else	  drawCell(room, x, y, 0);	
		}
		x++;
	}
}



jswByte *Jsw48RoomDraw::getRoom(int room)
{
	return m_game->getRoom(room);
}


int Jsw48RoomDraw::onReveal(int whichRectangle)
{
	if (!m_haveTimer)
	{
		m_ot = setTickListener(this);
		m_ticker = 0;
		m_haveTimer = 1;
	}
	return VideoNotePage::onReveal();
}







void Jsw48RoomDraw::intPreContextMenu(void)
{
        if (m_haveTimer)
        {
                setTickListener(m_ot);
                m_ot = NULL;
                m_haveTimer = 0;
        }
}


void Jsw48RoomDraw::postContextMenu(void)
{
	m_ot = setTickListener(this);	
	m_haveTimer = 1;
	VideoNotePage::postContextMenu();
}



int Jsw48RoomDraw::onConceal()
{
	if (m_haveTimer)
        {
                setTickListener(m_ot);
                m_ot = NULL;
                m_haveTimer = 0;
        }

	return VideoNotePage::onConceal();
}


int Jsw48RoomDraw::redraw(int whichRectangle)
{
	int rv = VideoNotePage::redraw(whichRectangle);
	if (whichRectangle & VPR_RIGHT)
	{
		drawFixed();
		drawData();
	}
	m_visible = 1;
	return rv;
}

void Jsw48RoomDraw::gotoRoom(int room)
{
	jswByte *b = m_game->getRoom(room);
	if (!b) return;

        if (m_bmAir)    delete m_bmAir;    m_bmAir    = NULL;
	if (m_bmWater)  delete m_bmWater;  m_bmWater  = NULL;
	if (m_bmEarth)  delete m_bmEarth;  m_bmEarth  = NULL;
        if (m_bmFire)   delete m_bmFire;   m_bmFire   = NULL;
        if (m_bmRamp)   delete m_bmRamp;   m_bmRamp   = NULL;
        if (m_bmConvey) delete m_bmConvey; m_bmConvey = NULL;
	if (m_bmObject) delete m_bmObject; m_bmObject = NULL;

	m_room = room;
	loadRoom(m_room);

	drawData();
}




void Jsw48RoomDraw::loadRoom(int room, int reload)
{
	int ng, n, bg;
	jswByte *b = m_game->getRoom(room);

	bg = (b[0xA0] >> 3) & 0x0F;
	b  += 0xF0;
	jswByte *p, *g = m_game->getGuards(room);

	for (n = 0; n < (8 + FG_MAX); n++)
	{
		if (m_guards[n]) delete m_guards[n];
		m_guards[n] = NULL;
	}
	for (n = 0; n < 8; n++)
	{	
		ng = (*b) & (m_game->getGuardianMask());
		if (ng == m_game->getGuardianMask()) break;
		++b;
		p = g + (8 * ng);
		if (*p == 0xFF) break;
		m_guards[n] = m_game->newGuardian(p, *b, bg);
		++b;	
	}
	checkSpecials();
}

void Jsw48RoomDraw::checkSpecials()
{
	int n = 8;
	int bg;

        jswByte *b = m_game->getRoom(m_room);
        bg = (b[0xA0] >> 3) & 0x0F;

	if (m_game->getStartRoom() == m_room)
	{
		m_guards[n++] = new JswFixedGuard(FG_WILLY, m_game, bg);
	}
	if (m_game->getMariaRoom() == m_room)
	{
		m_guards[n++] = new JswFixedGuard(FG_MARIA, m_game, bg);
	}
	if (m_game->getToiletRoom() == m_room)
	{
		m_guards[n++] = new JswFixedGuard(FG_TOILET, m_game, bg);
	}	

}


void Jsw48RoomDraw::drawObject(int ink, int x, int y)
{
	m_bmObject->getSurface()->fillReplace(0, 0, 16, 16, ZX_WHITE, ink);
	m_bmObject->toSurface(m_bmAnim, x*16, y*16); 
	m_bmObject->getSurface()->fillReplace(0, 0, 16, 16, ink, ZX_WHITE);
}

void Jsw48RoomDraw::drawObjects(void)
{
	int ink = 3 + ((m_ticker / 5) % 4);
	int n,x,y, rm;

	for (n = 0; n < 256; n++)
	{
		if ((!m_game->objectPos(n, &rm, &x, &y)) && rm == m_room)
		{
			drawObject(ink++, x, y/8);
	
			if (ink == 7) ink = 3;
		}
	}

}                                                     



void Jsw48RoomDraw::onTick(void)
{
	flashColours();
}



