/************************************************************************

    JSWED 2.00 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "j128guard.hxx"
#include "j128glist.hxx"
#include "guardedit.hxx"
#include "j128gedit.hxx"

JswGame::JswGame(SpectrumMemory *mem)
{
	m_upgrade = NULL;
	m_mem = mem;
	m_filename[0] = 0;
	m_objectXY    = 0xA500;
	m_objectCount = 0xA3FF;
	m_objectBase  = 0xA400;
	m_objectStat  = 0xA400;
	m_objectMask  = 0x3F;
	m_guardianMask= 0x7F;
	m_guardsBase  = 0xA000;
	m_guardsEnd   = 0xA600;
	m_spriteBase  = 0x9C00;

	/* Initial PC for JSW48 */
	m_mem->poke(0x5BFE, 0xCA);
	m_mem->poke(0x5BFF, 0x87);

}

JswGame::~JswGame()
{
	delete m_mem;
}

char *JswGame::getID()
{
	char *name, *pname;
	
	name  = m_filename;
	pname = strrchr(name, '/');
	if (pname) name = pname + 1;
#ifdef _WINDOWS
	pname = strrchr(name, '\\');
	if (pname) name = pname + 1;
	pname = strrchr(name, ':');
	if (pname) name = pname + 1;
#endif
	sprintf(m_namebuf, "%-40.40s", name);
	return m_namebuf;
}





int JswGame::editMain()
{
	m_keyFilter = setAppKeyFilter(this);
	JswNotebook p(0, 0, VIDEO_W, VIDEO_H);
	p.setTitle("Screens");
	addPages(&p);
	int rv = p.doModal();
	deletePages();
	setAppKeyFilter(m_keyFilter);
	return rv;
}


int JswGame::getStartRoom(void)
{
	return m_mem->peek(0x87EB);
}

void JswGame::setStartRoom(int room)
{
	m_mem->poke(0x87EB, room);
}

int JswGame::getMariaRoom(void)
{
        return m_mem->peek(0x9538);
}

void JswGame::setMariaRoom(int room)
{
        m_mem->poke(0x9538, room);
}

int JswGame::getToiletRoom(void)
{
        return m_mem->peek(0x959E);
}

void JswGame::setToiletRoom(int room)
{
        m_mem->poke(0x959E, room);
}





void JswGame::objectDelete(int index)
{
	int objmin = m_mem->peek(m_objectCount);

	if (objmin == 0xFF) return;	// Can't delete last object
	// Move all objects up one
	for (int n = index; n > objmin; n--)
	{
		m_mem->poke(m_objectBase + n, m_mem->peek(m_objectBase + n - 1));
		m_mem->poke(m_objectXY   + n, m_mem->peek(m_objectXY   + n - 1));
		m_mem->poke(m_objectStat + n, m_mem->peek(m_objectStat + n - 1));
	}
	m_mem->poke(m_objectCount, objmin + 1);
}


void JswGame::objectInsert(int room, int x, int y)
{
	int objmin = m_mem->peek(m_objectCount);
	
	if (!objmin) return;	// All objects taken

	--objmin;
	y *= 2;
	m_mem->poke(m_objectStat + objmin, ((y & 0x80) | 0x40)); 
	m_mem->poke(m_objectXY   + objmin, (x & 31) | ((y & 0x70) << 1));
	m_mem->poke(m_objectBase + objmin,
		  (m_mem->peek(m_objectBase + objmin) & ~m_objectMask) | room); 
	m_mem->poke(m_objectCount, objmin);
}


int  JswGame::objectPos(int index, int *room, int *x, int *y)
{
	int oy = m_mem->peek(m_objectStat + index) & 0x80;
	int ox = m_mem->peek(m_objectXY   + index);

	if (index < m_mem->peek(m_objectCount)) return -1;

	*room = m_mem->peek(m_objectBase + index) & m_objectMask;

	*x = (ox & 0x1F);
	*y = ((ox & 0xE0) >> 2)	| (oy >> 1); 
	return 0;
}


int  JswGame::objectAt(int room, int x, int y)
{
        int objmin = m_mem->peek(m_objectCount);
        int n, room1, x1, y1;

	for (n = objmin; n <= 0xFF; n++)
	{
		if (!objectPos(n, &room1, &x1, &y1))
		{
			if (x1 == x && y1 == y && room1 == room) 
				return n;
		}

	}
	return -1;
}

int JswGame::guardTableCount(void)
{
	return 1;
}


int  JswGame::chooseGuardTable(int *ng)
{
	*ng = m_guardsBase;
	return ZXE_OK;
}

jswByte *JswGame::getGuards(int room)
{
        return m_mem->memoryAt(m_guardsBase);
}

int JswGame::chooseGuard(VideoBitmap *bm, jswByte *table, int *ng, int bg)
{
	Jsw128GuardList gl(bm, table, m_mem, bg, getGuardianMask());

	int rv = gl.doModal();

	if (rv >= ZXE_CANCEL) return rv;
	*ng = gl.getSelected();	
	return rv;
}

int JswGame::allowSuperJump(void)
{
	return 0;
}


int JswGame::allowSpriteOverride(void)
{
	if (m_mem->peek(0x9654) == 0x20 &&
            m_mem->peek(0x9655) == 0x84) return 0;	
	return m_mem->peek(0x9654);
}


int JswGame::decodeWillySprite(int room, int sprite)
{
	int j = allowSpriteOverride();
	
	if (j) return sprite;
	
	if (room == m_mem->peek(0x9657)) return m_mem->peek(0x965B);
	else				 return m_mem->peek(0x9652);
}

int JswGame::encodeWillySprite(int room, int sprite)
{
	int j = allowSpriteOverride();

	if (j) return sprite;

	return 0;
}



void JswGame::getStartPos(int *x, int *y)
{
	*x = m_mem->peek(0x87F0) & 0x1F;
	*y = m_mem->peek(0x87e6);
}


void JswGame::setStartPos(int x, int y)
{
	m_mem->poke(0x87F0, ((y << 1) & 0xE0) | (x & 0x1F));
	m_mem->poke(0x87F1, (m_mem->peek(0x87F1) & 0xFE) | (y >> 7));
	m_mem->poke(0x87E6, y);
}

void JswGame::getMariaPos(int *x, int *y)
{
	int hl = m_mem->peek(0x955F) + 256 * m_mem->peek(0x9560);
	int y1, y2, y3;

	*x =  hl & 0x1F;
	
	hl &= 0x0FE0;
        y1 = (hl >>  8) & 7;
        y2 = (hl >>  5) & 7;
        y3 = (hl >> 11) & 3;

        *y = 2*(y1 + 8*y2 + 64*y3);
}

void JswGame::setMariaPos(int x, int y)
{
	int hl = m_mem->peek(0x955F) + 256 * m_mem->peek(0x9560);
	int y1, y2, y3;
	y &= 0xF0;
	int w = (2*y) + x;
	
	hl = (hl & ~(0xFFF)) | x;
	y1 = ((y >> 1) & 7) << 8;
	y2 = ((y >> 4) & 7) << 5; 
	y3 = ((y >> 7) & 3) << 11;
	hl |= (y1+y2+y3);

	m_mem->poke(0x955F, (hl & 0xFF));
	m_mem->poke(0x9560, (hl >> 8));

	hl = 0x5C00 + w;
        m_mem->poke(0x956d, (hl & 0xFF));
        m_mem->poke(0x956e, (hl >> 8));
	hl += 32;
        m_mem->poke(0x9573, (hl & 0xFF));
        m_mem->poke(0x9574, (hl >> 8));
}


void JswGame::getToiletPos(int *x, int *y)
{	
	*x = m_mem->peek(0x95B9) & 0x1F;
	*y = m_mem->peek(0x95B6);
}

void JswGame::setToiletPos(int x, int y)
{
        int hl;

	m_mem->poke(0x95B9, x);
	m_mem->poke(0x95B6, y);

	y *= 2;
	y &= 0x1E0;
	m_mem->poke(0x958E, x | (y & 0xE0));

        hl = 0x5C00 + y + x;
        m_mem->poke(0x95c2, (hl & 0xFF));
        m_mem->poke(0x95c3, (hl >> 8));
        hl += 32;
        m_mem->poke(0x95c5, (hl & 0xFF));
        m_mem->poke(0x95c6, (hl >> 8));

}


int JswGame::prevRoom(int room)
{
	int o = room;
	do
	{
		room--;
		if (room < 0) return o;
	} while (roomReserved(room));
	return room;
}

int JswGame::nextRoom(int room)
{
	int o = room;
	do
	{
		room++;
		if (room > 255) return o;
	} while (roomReserved(room));
	return room;
}




JswRoomList *JswGame::getRoomList(void)
{
        int n;
        jswByte *b;
        char rname[40];

        JswRoomList *s = new JswRoomList();

        for (n = 0; n < 256; n++) if (!roomReserved(n))
        {
                b = getRoom(n);
                sprintf(rname, "%03d %-32.32s", n, b + 128);
                s->addString(rname);
        } 
        return s;
}


int JswGame::canTeleport(void)
{
	if (m_mem->peek(35696) == 205) return 1;
	return 0;	
}



jswByte *JswGame::getTeleportList(void)
{
	int addr;

	if (!canTeleport()) return NULL;

	addr = m_mem->peek(35698) * 256 + m_mem->peek(35697);
	addr = m_mem->peek(addr+6) * 256 + m_mem->peek(addr+5);

	return m_mem->memoryAt(addr);
}

int JswGame::teleUnload(void)
{
	m_mem->poke(35696, 0x01);
	m_mem->poke(35697, 0xFE);
	m_mem->poke(35698, 0xEF);
	return ZXE_CONTINUE;
}

int JswGame::teleLoad(int page)
{
	static jswByte tport[] = {
        	0xD5,0xDD,0xE5,0xDD,0x21,0x4A,0x01,0x11,
	        0x04,0x00,0xDD,0x46,0x00,0xDD,0x23,0x78,
       		0xB7,0x28,0x30,0x3A,0x20,0x84,0xDD,0xBE,
		0x00,0x20,0x24,0x3A,0xCF,0x85,0xDD,0xBE,
	        0x01,0x20,0x1C,0x3A,0xD3,0x85,0xE6,0x1F,
		0xDD,0xBE,0x02,0x20,0x12,0xDD,0x7E,0x03,
		0x32,0x20,0x84,0xE5,0x00,0x00,0x00,0xE1,
		0xDD,0xE1,0xD1,0xC1,0xC3,0x12,0x89,0xDD,
		0x19,0x10,0xD0,0xDD,0xE1,0xD1,0x01,0xFE,
		0xEF,0xC9,0x00 };

	memcpy(m_mem->memoryAt(page * 256), tport, 
		sizeof(tport));
	m_mem->poke(page * 256 + 6, page);
        m_mem->poke(35696, 0xCD);
        m_mem->poke(35697, 0x00);
        m_mem->poke(35698, page);
	return ZXE_CONTINUE;
}

int JswGame::teleLoad(void)
{
	int rv;

	JswRoomList *sl = getRoomList();	

	if(!sl) return ZXE_CONTINUE;

	sl->setTitle("Room to be replaced");
	rv = sl->doModal();	

	if (rv >= ZXE_QUIT) { delete sl; return rv; }
	if (rv != ZXE_OK)   { delete sl; return ZXE_CONTINUE; }

	teleLoad(sl->getSelectedRoom() + 0xC0);
	delete sl;

	return ZXE_CONTINUE;
}



int JswGame::toggleTeleporters(void)
{
	int rv;

	if (canTeleport())
	{
		VideoMenu vm("Delete all teleporters?", "Confirm", "Cancel",
				NULL);

		rv = vm.doModal(); 
		if (rv >= ZXE_QUIT) return rv;
		if (rv != ZXE_OK || vm.getSelected() != 0) return ZXE_CONTINUE;
		teleUnload();
		return rv;	
	}
	return teleLoad();
}



int JswGame::guardianEditor(int guardian, jswByte *table, int x, VideoBitmap *backg)
{
	jswByte theGuard[8];

	memcpy(theGuard, table + 8 * guardian, 8);

	GuardianEditor *g = newGuardianEditor(theGuard, backg, x);

	int rv = g->doModal();
	delete g;
	
	if (rv >= ZXE_QUIT) return rv;
	if (rv == ZXE_OK) memcpy(table + 8 * guardian, theGuard, 8);
	return rv;
}

GuardianEditor *JswGame::newGuardianEditor(jswByte *guard, VideoBitmap *b, int x)
{
	return new Jsw128GuardEditor(this, guard, b, x);
}


int JswGame::supportsExtGuards(void)
{
	return 0;
}


jswByte *JswGame::getFont(int *len)
{
	if (len) *len = 0; 
	return NULL;
}



int JswGame::clearRooms(void)
{
	jswByte rm[256];
	jswByte *b;
	VideoMenu vm("This is destructive!", 
                     "Delete all rooms", "Cancel", NULL);

	int rv = vm.doModal();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK || vm.getSelected() != 0) return ZXE_CONTINUE;

	memset(rm, 0, sizeof(rm));
	rm[0xF0] = 0xFF;
	for (int n = 0; n < 256; n++)
	{			 
		sprintf((char *)rm + 128, "           Room %03d             ",n);
		b = getRoom(n);
		if (b) memcpy(b, rm, 256);	
	}
	// Delete all items
	m_mem->poke(m_objectCount, 0xFF);
	return ZXE_CONTINUE;
}


int JswGame::clearSprites(void)
{
       VideoMenu vm("This is destructive!",
                     "Delete all sprites", "Cancel", NULL);

        int rv = vm.doModal();
        if (rv >= ZXE_QUIT) return rv;
        if (rv != ZXE_OK || vm.getSelected() != 0) return ZXE_CONTINUE;

	deleteSprites();
	return ZXE_CONTINUE;
}


int JswGame::clearGuardians(void)
{
	char tbuf[40];
	int gt;
	int rv  = chooseGuardTable(&gt);
	jswByte *b;
	
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK)   return ZXE_CONTINUE;

	sprintf(tbuf, "Clear guardian table %04x", gt);
        VideoMenu vm("This is destructive!",
                     tbuf, "Cancel", NULL);
	rv = vm.doModal();
        if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK)   return ZXE_CONTINUE;

	b = m_mem->memoryAt(gt);
	memset(b, 0, 1023);
	b[1016] = 0xFF;	
	return ZXE_CONTINUE;
}

int JswGame::onKeyDown(int keysym)
{
	switch(keysym)
	{
		case ZXK_F5:	return onTest();
		case ZXK_F2:	return onSave();
	}

        if (m_keyFilter) return m_keyFilter->onKeyDown(keysym);
        return -1;
}


int JswGame::onKeyUp(int keysym)
{
	if (m_keyFilter) return m_keyFilter->onKeyUp(keysym);
	return -1;
}


int JswGame::onTest(void)
{
	char temp_name[PATH_MAX];
	FILE *fp = openTemp(temp_name, ".z80", "w+b");
	if (!fp) 
	{
		alert("Couldn't save temp file", "Cancel");
		return ZXE_CONTINUE;
	}
	m_mem->saveZ80(fp);
	fclose(fp);

	return executeZ80(temp_name);
}

int JswGame::onSave(void)
{
        int rv;

	if (!m_filename[0]) 
	{
		return alert("No filename, so can't save ", "Cancel");
	}
        FILE *fp = fopen(m_filename, "wb");
        if (!fp)
        {
		return alert("Failed to open file", "Cancel");
        }
        SpectrumMemory *mem = getMem();

        rv = mem->toDisc(m_filename, fp);
        fclose(fp);
        if (rv >= ZXE_QUIT) return rv;

	rv = alert("Game saved", "OK"); 
	if (rv >= ZXE_QUIT) return rv;
	return ZXE_CONTINUE;
}

int JswGame::getObjectCount()
{
	return m_mem->peek(m_objectCount);
}

jswByte *JswGame::getTitleAttrs(void)
{
        return m_mem->memoryAt(0x9800);
}        


jswByte JswGame::getWillySpritePage(void) 
{
	return 0x9D;
}


Guardian *JswGame::newGuardian(jswByte *guard, int x, int bg)
{
	return new Jsw128Guard(guard, x, m_mem, bg);
}

