/************************************************************************

    JSWED 2.00 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "screenedit.hxx"
#include "tick.xbm"
#include "pencil.xbm"

#define MD_BITMAP 0
#define MD_ATTRIB 1

#define GR_NONE  0
#define GR_PIXEL 1
#define GR_ATTR  2

#define GRID_C1  0x28
#define GRID_C2  0x79

ScreenEditorPage::ScreenEditorPage()
{
	int x, y, w, h;

	m_bDraw = false;
	m_bmData = m_bmAttr1 = m_bmAttr2 = m_bmPixGrid = NULL;
	m_tcount = 0;
	m_haveTimer = 0;
	m_ot = NULL;
	m_pcx = m_pcy = 0;
	m_acx = m_acy = 0;
	m_visible = 0;
	m_attrib = 0x47;
	m_buttonDown = 0;
	m_grid = GR_NONE;	
	m_mode = MD_BITMAP;
	m_undoBits = NULL;
	m_undoAttrs = NULL;

	m_bmTick = bitmapFromXbm(tick_width, tick_height, tick_bits, 0,
				ZX_BRBLUE, ZX_BRWHITE);
	m_bmPencil = bitmapFromXbm(pencil_width, pencil_height, pencil_bits, 0,
				ZX_BRBLUE, ZX_BRWHITE);
	m_bmColours = newVideoBitmap(16, 16);
	for (y = 0; y < 16; y++) for (x = 0; x < 16; x++)
	{
		m_bmColours->getSurface()->plot(x, y, (x^y));
	}
	m_bmCgrid = newVideoBitmap(16,16);
	m_bmPgrid = newVideoBitmap(16,16);
	for (y = 0; y < 16; y += 4) for (x = 0; x < 16; x += 4) 
	{
		int c = ((x ^ y) & 4) ? ZX_BRWHITE : ZX_CYAN;
		m_bmCgrid->getSurface()->fillBox(x, y, 4, 4, c);

		m_bmPgrid->getSurface()->fillBox(x, y, 4, 4, ZX_BRWHITE);
		m_bmPgrid->getSurface()->plot(x, y, ZX_BLACK);
	}

        x = VIDEO_W - 16 * CHAR_W;
        y = 24 * CHAR_H + CHAR_SH;
	w = 4 * CHAR_W;
	h = 2 * CHAR_H;

	addHotspot(new Hotspot('I', x, y, w, h, 'I', 'i')); x += w; 
	addHotspot(new Hotspot('P', x, y, w, h, 'P', 'p')); x += w;
	addHotspot(new Hotspot('B', x, y, w, h, 'B', 'b')); x += w; 
	addHotspot(new Hotspot('F', x, y, w, h, 'F', 'f'));

	x = 0;
	addHotspot(new Hotspot('M', x, y, w, h, 'M', 'm'));

	x = 5 * CHAR_W;
	addHotspot(new Hotspot('G', x, y, w, h, 'G', 'g'));
	
	x = 10 * CHAR_W;
	w /= 2;
	addHotspot(new Hotspot('<', x, y, w, h, '<')); x += w;
	addHotspot(new Hotspot('>', x, y, w, h, '>'));
}


ScreenEditorPage::~ScreenEditorPage()
{
	deleteBitmaps();	
	delete m_bmTick;
	delete m_bmColours;
	delete m_bmPgrid;
	delete m_bmCgrid;
	delete m_bmPencil;
	if (m_haveTimer) setTickListener(m_ot);
}

void ScreenEditorPage::deleteBitmaps(bool withUndo)
{
	if (withUndo)
	{
		if (m_undoBits)  delete m_undoBits;
		if (m_undoAttrs) delete m_undoAttrs;
		m_undoBits = m_undoAttrs = NULL;
	}
	if (m_bmData)    delete m_bmData;
	if (m_bmAttr1)   delete m_bmAttr1;
	if (m_bmAttr2)   delete m_bmAttr2;
	if (m_bmPixGrid) delete m_bmPixGrid;
        m_bmData = m_bmAttr1 = m_bmAttr2 = m_bmPixGrid = NULL;
}

void ScreenEditorPage::setup(bool withUndo)
{	
	int x, y;

	deleteBitmaps(withUndo);

	switch(m_format)
	{
		case FMT_XBM:   m_mode   = MD_BITMAP;
				m_bmData = bitmapFromXbm(m_w, m_h, m_bits, 
					1, ZX_ORANGE, ZX_BRORANGE); break;
		case FMT_ROSCRN:
		case FMT_SCREEN: m_bmData = bitmapFromScreen(m_h, m_bits,
					1, ZX_ORANGE, ZX_BRORANGE); break;
		case FMT_SPRITE: m_bmData = bitmapFromSprite(m_w, m_h, m_bits, 
					1, ZX_ORANGE, ZX_BRORANGE); break;
	}
	m_bmPixGrid = newVideoBitmap(m_w * 2, m_h * 2);
	m_bmPixGrid->getSurface()->fillBox(0, 0, m_w * 2, m_h * 2, ZX_BRORANGE);
	for(x = 6; x < 2*m_w; x += 16) for (y = 6; y < 2*m_h; y+= 16)
	{
		m_bmPixGrid->getSurface()->box(x, y, 4, 4, ZX_ORANGE);
	} 
	if (withUndo)
	{
		m_undoBits = new jswByte[m_w * m_h / 8];	
		memcpy(m_undoBits, m_bits, m_w * m_h / 8);
	}
	m_bmAttr1 = newVideoBitmap(m_w*2, m_h*2);
        m_bmAttr2 = newVideoBitmap(m_w*2, m_h*2);

	m_aw = (m_w + 7) / 8;	
	m_ah = (m_h + 7) / 8;

	if (m_acx >= m_aw) m_acx = m_aw - 1;
	if (m_acy >= m_ah) m_acy = m_ah - 1;
	if (m_pcx >= m_w)  m_pcx = m_w - 1;
	if (m_pcy >= m_h)  m_pcy = m_h - 1;

	if (m_attrs)
	{
		for (x = 0; x < m_aw; x++) for (y = 0; y < m_ah; y++) 
		{
			setCellAttribute(x, y, m_attrs[y*m_aw + x]);
		}
		if (withUndo)
		{
			m_undoAttrs = new jswByte[m_aw * m_ah];
			memcpy(m_undoAttrs, m_attrs, m_aw * m_ah);
		}
	}
	else
	{
		for (x = 0; x < m_aw; x++) for (y = 0; y < m_ah; y++)
		{
			setCellAttribute(x, y, 0x46);
		}
		if (withUndo) m_undoAttrs = NULL;
	}
	if (m_format == FMT_XBM || m_format == FMT_ROSCRN)
		m_mode = MD_ATTRIB;
	if (!m_attrs) m_mode = MD_BITMAP;
}


int ScreenEditorPage::redraw(int whichRectangle)
{
	m_visible = 1;
	int rv = VideoNotePage::redraw(whichRectangle);

	int x,y,w,h;
	x = VIDEO_W - 16 * CHAR_W;
	y = 24 * CHAR_H;
	w = 4 * CHAR_W;
	h = 2 * CHAR_H;
	videoScreen->fillBox(x, y, 16 * CHAR_W, 5*CHAR_SH, ZX_BRWHITE);
	videoScreen->box(x, y + CHAR_SH, 16 * CHAR_W, h, ZX_BLACK);
	videoScreen->drawSmallText(x, y, "Attributes                      ", ZX_BRWHITE, ZX_BLACK);
	videoScreen->drawText(x + CHAR_SW, y + CHAR_H, "I:* P:* B:* F:"
			, ZX_BLACK, ZX_BRWHITE);
	videoScreen->smallZxLogo(VIDEO_W - 6 * CHAR_SW, y);
	videoScreen->box(x, y + CHAR_SH, w, h, ZX_BLACK);
	videoScreen->box(x, y + CHAR_SH, w + 4*CHAR_W, h, ZX_BLACK);
	videoScreen->box(x, y + CHAR_SH, w + 8*CHAR_W, h, ZX_BLACK); 

	x = 0;
	videoScreen->fillBox(x, y, w, 5 * CHAR_SH, ZX_BRWHITE);
	videoScreen->box(x, y + CHAR_SH, w, h, ZX_BLACK);
	videoScreen->drawSmallText(x, y, "Mode    ", ZX_BRWHITE, ZX_BLACK);
	videoScreen->drawText(x + CHAR_SW, y + CHAR_H, "M:*", ZX_BLACK, ZX_BRWHITE);

	x = 5 * CHAR_W;
        videoScreen->fillBox(x, y, w, 5 * CHAR_SH, ZX_BRWHITE);
        videoScreen->box(x, y + CHAR_SH, w, h, ZX_BLACK);
        videoScreen->drawSmallText(x, y, "Grid    ", ZX_BRWHITE, ZX_BLACK);
        videoScreen->drawText(x + CHAR_SW, y + CHAR_H, "G:*", ZX_BLACK, ZX_BRWHITE);

	x = 10 * CHAR_W;
        videoScreen->fillBox(x, y, w, 5 * CHAR_SH, ZX_BRWHITE);
        videoScreen->box(x, y + CHAR_SH, w, h, ZX_BLACK);
        videoScreen->drawSmallText(x, y, "Go to   ", ZX_BRWHITE, ZX_BLACK);
        videoScreen->drawText(x + CHAR_SW, y + CHAR_H, "< >", ZX_BLACK, ZX_BRWHITE);
	videoScreen->box(x, y+ CHAR_SH, w/2, h, ZX_BLACK);

	if (m_h < 192)
	{
		videoScreen->fillBox(m_x, m_y + 2 * m_h, 512, 384 - 2 * m_h, ZX_WHITE);
	}
	if (m_w < 256)
	{
                videoScreen->fillBox(m_x + 2 * m_w, 0, 512 - 2 * m_w, 384, ZX_WHITE);
	}

	drawData(0);

	return rv;
}


void ScreenEditorPage::drawData(int flashOnly)
{
	int t = m_tcount % 40;

	if (t >= 20 && m_bmAttr2) m_bmAttr2->toScreen(m_x, m_y);
        else if       (m_bmAttr1) m_bmAttr1->toScreen(m_x, m_y);

	drawCursor();
	
	if (!flashOnly)
	{
		showMode();
		showGrid();
		showAttributes();
		drawZoom();
	}
	else drawZoomCursor(videoScreen);

}

void ScreenEditorPage::drawCursor(void)
{
	int t = m_tcount % 40;
	int ci;

	if (t >= 20) ci = ZX_BRBLUE;
	else	     ci = ZX_BRWHITE;

	if (m_mode == MD_ATTRIB) 
	{
		videoScreen->box(m_x + 16 * m_acx, m_y + 16 * m_acy, 16, 16, ci);	
	}
	else if (t < 10 || t >= 30)
	{
		drawPixel(m_pcx, m_pcy);
		videoScreen->plot(m_x + 2 * m_pcx,   m_y + 2*m_pcy,   ci);
		videoScreen->plot(m_x + 2 * m_pcx+1, m_y + 2*m_pcy+1, ci);
	}
	else
	{
		drawPixel(m_pcx, m_pcy);
                videoScreen->plot(m_x + 2 * m_pcx+1, m_y + 2*m_pcy,   ci);
                videoScreen->plot(m_x + 2 * m_pcx,   m_y + 2*m_pcy+1, ci);
	}
}


void ScreenEditorPage::drawCell(int x, int y)
{
        int t = m_tcount % 40;
	int sx = x * 16;
	int sy = y * 16;

        if (t >= 20 && m_bmAttr2) m_bmAttr2->toScreen(m_x+sx, m_y + sy,
				    sx, sy, 16, 16);
        else if       (m_bmAttr1) m_bmAttr1->toScreen(m_x+sx, m_y + sy,
				    sx, sy, 16, 16);
}

void ScreenEditorPage::drawPixel(int x, int y)
{
        int t = m_tcount % 40;
        int sx = x * 2;
        int sy = y * 2;

        if (t >= 20 && m_bmAttr2) m_bmAttr2->toScreen(m_x+sx, m_y + sy,
                                    sx, sy, 2, 2);
        else if       (m_bmAttr1) m_bmAttr1->toScreen(m_x+sx, m_y + sy,
                                    sx, sy, 2, 2);

}

void ScreenEditorPage::showMode(void)
{
        int x = CHAR_SW + 2 * CHAR_W;
        int y = 25 * CHAR_H;

	switch(m_mode)
	{
		case MD_BITMAP: m_bmPencil->toScreen(x, y); break;
		case MD_ATTRIB: m_bmColours->toScreen(x, y); break;	
	}
}

void ScreenEditorPage::showGrid(void)
{
	int x = 7 * CHAR_W + CHAR_SW;
	int y = 25 * CHAR_H;

	switch(m_grid)
	{
		case GR_NONE:	videoScreen->fillBox(x, y, 16, 16, ZX_BRWHITE);
				break;
		case GR_ATTR:	m_bmCgrid->toScreen(x, y);  break;
		case GR_PIXEL:	m_bmPgrid->toScreen(x, y);  break;
	}
}


//
// Show the currently selected attributes
//
void ScreenEditorPage::showAttributes(void)
{
        int x,y;
	int ifg, ibg, pfg, pbg;

	ibg = (m_attrib & 7) | ZX_BRIGHT;		// Current ink
	pbg = ((m_attrib >> 3) & 7) | ZX_BRIGHT;	// Current paper

	//
	// "INK 9" - choose black/white ink for contrast
	//
	if (ibg >= ZX_BRGREEN) ifg = ZX_BLACK; else ifg = ZX_BRWHITE;
	if (pbg >= ZX_BRGREEN) pfg = ZX_BLACK; else pfg = ZX_BRWHITE;

        x = VIDEO_W - 14 * CHAR_W + CHAR_SW;
        y = 25 * CHAR_H;

	//
	// Display colours
	//
	videoScreen->drawText(x,          y, (ibg + '0' - ZX_BRIGHT), ifg, ibg);
	videoScreen->drawText(x+4*CHAR_W, y, (pbg + '0' - ZX_BRIGHT), pfg, pbg);

	//
	// Empty the bright/flash checkboxes
	//
	videoScreen->fillBox(x + 8*CHAR_W, y, CHAR_W, CHAR_H, ZX_BRWHITE);
        videoScreen->fillBox(x +12*CHAR_W, y, CHAR_W, CHAR_H, ZX_BRWHITE);

	//
	// And set them as appropriate.
	//
	if (m_attrib & 0x40) m_bmTick->toScreen(x +  8 * CHAR_W + 2, y+2);
        if (m_attrib & 0x80) m_bmTick->toScreen(x + 12 * CHAR_W + 2, y+2);
	
}

void ScreenEditorPage::onTick(void)
{
	if (!m_visible) return;
	++m_tcount;
	if ((m_tcount % 20) == 0)
	{
		drawData(1);
	}
	else if ((m_tcount % 20) == 10) 
	{
		drawCursor();
		drawZoomCursor(videoScreen);
	}

}


int ScreenEditorPage::preContextMenu(void)
{
         if (m_haveTimer)
        {
                setTickListener(m_ot);
                m_tcount = 0;
                m_haveTimer = 0;
        }
	return VideoNotePage::preContextMenu();
}


void ScreenEditorPage::postContextMenu(void)
{
        if (!m_haveTimer)
        {
                m_ot = setTickListener(this);
                m_tcount = 0;
                m_haveTimer = 1;
        }
	VideoNotePage::postContextMenu();
}

void ScreenEditorPage::setCellAttribute(int x, int y, jswByte b)
{
	//
	// If there is an attribute grid, then the colours are fixed
	//
	if (m_grid == GR_ATTR) 
	{
		if ((x ^ y) & 1) b = GRID_C1;
		else		 b = GRID_C2; 
	}

	jswByte fg = (b & 7), bg = (b >> 3) & 7;

	if (b & 64) { fg |= ZX_BRIGHT; bg |= ZX_BRIGHT; }

	VideoBitmap *bm = (m_grid == GR_PIXEL) ? m_bmPixGrid : m_bmData;

        bm->toSurface(m_bmAttr1->getSurface(), x*16, y*16, x*16, y*16, 16, 16);
        bm->toSurface(m_bmAttr2->getSurface(), x*16, y*16, x*16, y*16, 16, 16);

	m_bmAttr1->getSurface()->fillReplace(x*16,y*16,16,16,ZX_ORANGE,  fg);
	m_bmAttr1->getSurface()->fillReplace(x*16,y*16,16,16,ZX_BRORANGE,bg);

	if (b & 128) { b = fg; fg = bg; bg = b; } 

        m_bmAttr2->getSurface()->fillReplace(x*16,y*16,16,16,ZX_ORANGE,  fg);
        m_bmAttr2->getSurface()->fillReplace(x*16,y*16,16,16,ZX_BRORANGE,bg);
}


void ScreenEditorPage::togglePixel(int x, int y)
{
	bool b = getPixel(x, y);
	setPixel(x, y, !b);
}

int ScreenEditorPage::onReveal(int whichRectangle)
{
        if (!m_haveTimer)
        {
                m_ot = setTickListener(this);
                m_tcount = 0;
                m_haveTimer = 1;
        }
        return VideoNotePage::onReveal();
}


int ScreenEditorPage::onConceal(void)
{
        if (m_haveTimer)
        {
		setTickListener(m_ot);
		m_tcount = 0;
                m_haveTimer = 0;
        }
        return VideoNotePage::onConceal();
}


int  ScreenEditorPage::onKeyDown(int keysym)
{
	int rv = onKey(keysym);
	if (rv != -1) return rv;

	if (keysym == 'Z' - '@' || keysym == 'U' - '@') return onUndo();

	if (m_mode == MD_ATTRIB) switch(keysym)
	{
		case ZXK_UP:    moveCursor(m_acx, m_acy - 1); return ZXE_CONTINUE;
                case ZXK_DOWN:  moveCursor(m_acx, m_acy + 1); return ZXE_CONTINUE;
                case ZXK_LEFT:  moveCursor(m_acx - 1, m_acy); return ZXE_CONTINUE;
                case ZXK_RIGHT: moveCursor(m_acx + 1, m_acy); return ZXE_CONTINUE;
		case ' ':	setCellAttribute(m_acx, m_acy, m_attrib);
       				m_attrs[m_acy*m_aw + m_acx] = m_attrib;
				drawCell(m_acx, m_acy);
			        drawCursor();
				drawZoomCursor(videoScreen);
				break;

	}
	else switch(keysym)
	{
                case ZXK_UP:    moveCursor(m_pcx, m_pcy - 1); return ZXE_CONTINUE;
                case ZXK_DOWN:  moveCursor(m_pcx, m_pcy + 1); return ZXE_CONTINUE;
                case ZXK_LEFT:  moveCursor(m_pcx - 1, m_pcy); return ZXE_CONTINUE;
                case ZXK_RIGHT: moveCursor(m_pcx + 1, m_pcy); return ZXE_CONTINUE;
                case ' ':       togglePixel(m_pcx, m_pcy);
                                drawPixel(m_pcx, m_pcy);
                                drawCursor();
				drawZoomCursor(videoScreen);
                                break;
	}
	return VideoNotePage::onKeyDown(keysym);
}

int ScreenEditorPage::onMouseMove(int x, int y, int xrel, int yrel)
{
	if (x >= m_x && x < (m_x + (16*m_aw)) &&
	    y >= m_y && y < (m_y + (16*m_ah)))
	{
		if (m_mode == MD_ATTRIB)
		{
			x = (x - m_x) / 16;
			y = (y - m_y) / 16;
			if (m_buttonDown)
			{
				setCellAttribute(x, y, m_attrib);
				m_attrs[m_acy*m_aw + m_acx] = m_attrib;
				drawCell(x, y);		
			}
			moveCursor(x, y);
			return ZXE_CONTINUE;		
		}
		else
		{
			x = (x - m_x) / 2;
			y = (y - m_y) / 2;
			if (m_buttonDown)
			{	
				setPixel(x, y, m_bDraw);
			}
			moveCursor(x, y);
			return ZXE_CONTINUE;
		}
	}
	return VideoNotePage::onMouseMove(x, y, xrel, yrel);
}



int ScreenEditorPage::onButtonDown(int x, int y, int button)
{
        int rv = onButton(x, y, button);
        if (rv != -1) return rv;

	if (x >= m_x && y >= m_y && x < (m_x + 2 * m_w) && y < (m_y + 2*m_h))
	{
		if (m_mode == MD_ATTRIB)
		{
			m_buttonDown = 1;
			setCellAttribute(m_acx, m_acy, m_attrib);
                        m_attrs[m_acy*m_aw + m_acx] = m_attrib;
			drawCell(m_acx, m_acy);
			drawCursor();
			drawZoomCursor(videoScreen);
		}
		else
		{
			m_buttonDown = 1;
			m_bDraw = !getPixel(m_pcx, m_pcy);
			setPixel(m_pcx, m_pcy, m_bDraw);
			drawCursor();
			drawZoom();
		}
	}
	return VideoNotePage::onButtonDown(x, y, button);
}


int ScreenEditorPage::onButtonUp  (int x, int y, int button)
{
	m_buttonDown = 0;
	return VideoNotePage::onButtonUp(x, y, button);
}



void ScreenEditorPage::moveCursor(int x, int y)
{
	if (m_mode == MD_ATTRIB)
	{
		if (x < 0 || y < 0 || x >= m_aw || y >= m_ah) return;

		drawCell(m_acx, m_acy);
		m_acx = x;
		m_acy = y;
		drawZoom();
	}
	else
	{
		if (x < 0 || y < 0 || x >= m_w || y >= m_h) return;
		drawPixel(m_pcx, m_pcy);
		m_pcx = x;
		m_pcy = y;
		drawZoom();
	}
	drawCursor();	
}

void ScreenEditorPage::drawZoom(void)
{
	int x, y;
	int x1, y1;
	if (m_mode == MD_ATTRIB) { x = m_acx;     y = m_acy;     }
	else			 { x = m_pcx / 8; y = m_pcy / 8; } 

	VideoBitmap *bm = newVideoBitmap(128, 128);
	for (x1 = 0; x1 < 8; x1++) for (y1 = 0; y1 < 8; y1++)	
	{
		int pixel = m_bmAttr1->getSurface()->pointZx(16 * x + 2*x1, 
							     16 * y + 2*y1);

		bm->getSurface()->fillBox(16 * x1, 16 * y1, 16, 16, pixel);
	}
	drawZoomCursor(bm->getSurface());
	bm->toScreen(m_lx, m_ly);
	delete bm;
}

void ScreenEditorPage::drawZoomCursor(VideoSurface *s)
{
	
	if (m_mode != MD_ATTRIB)
	{
		int t = m_tcount % 40;
		int ci, x1, y1;

	        if (t >= 20) ci = ZX_BRBLUE;
		else         ci = ZX_BRWHITE;

		x1 = (m_pcx % 8); y1 = (m_pcy % 8);
                s->box(m_lx + 16 * x1, m_ly + 16 * y1, 16, 16, ci);
	}
}

int ScreenEditorPage::onHotspot(int id)
{
	int oa = m_attrib;

	switch(id)
	{
		case 'I':
			m_attrib &= (~7);
			m_attrib |= (oa + 1) & 7;
			showAttributes();
			break;
		case 'P':
			m_attrib &= (~0x38);
			m_attrib |= (oa + 8) & 0x38;
			showAttributes();
			break;
		case 'B': 
			m_attrib ^= 0x40;
			showAttributes();
			break;
		case 'F':
			m_attrib ^= 0x80;
			showAttributes();
			break;
		case 'M': return changeMode();
		case 'G': return changeGrid();
		case '<': return onPrev();
		case '>': return onNext();
	}
	return ZXE_CONTINUE;
}


jswByte* ScreenEditorPage::lookupPixel(int x, int y, jswByte *mask)
{
	switch(m_format)
	{
		case FMT_ROSCRN:
                case FMT_SCREEN:
                *mask = 0x80 >> (x & 7);

                return m_bits + (x / 8) + ((y & (~63)) * 32) + ((y & 56)*4) 
				+ (y & 7) * 256;

                case FMT_SPRITE:
                *mask = 0x80 >> (x & 7);
                return m_bits + (y * ((m_w + 7)/8)) + (x/8);

		case FMT_XBM:
		*mask = 1 << (x & 7);
		return m_bits + (y * ((m_w + 7)/8)) + (x/8);

	}
	return NULL;
}


bool ScreenEditorPage::getPixel(int x, int y)
{
	jswByte mask;
	jswByte *b   = lookupPixel(x, y, &mask);

	return ((*b) & mask) ? true : false;
}


void ScreenEditorPage::setPixel(int x, int y, bool draw)
{
	// Read-only bitmaps
	if (m_format == FMT_XBM || m_format == FMT_ROSCRN) return;
	
	jswByte mask;
        jswByte *b   = lookupPixel(x, y, &mask);

	if (draw) *b |= mask;
	else	  *b &= ~mask;	

	m_bmData->getSurface()->fillBox(x*2,y*2, 2, 2, 
				(draw ? ZX_ORANGE : ZX_BRORANGE));

	if (m_attrs) mask = m_attrs[(y/8)*m_aw + (x/8)];
	else         mask = 0x46;

        //
        //        // If there is an attribute grid, then the colours are fixed

	if (m_grid == GR_ATTR) 
        {
        	if ((x ^ y) & 8) mask = GRID_C1;
        	else             mask = GRID_C2;
	}
        jswByte fg = (mask & 7), bg = (mask >> 3) & 7;

        if (mask & 64) { fg |= ZX_BRIGHT; bg |= ZX_BRIGHT; }

        VideoBitmap *bm = (m_grid == GR_PIXEL) ? m_bmPixGrid : m_bmData;

        bm->toSurface(m_bmAttr1->getSurface(), x*2, y*2, x*2, y*2, 2, 2);
        bm->toSurface(m_bmAttr2->getSurface(), x*2, y*2, x*2, y*2, 2, 2);

        m_bmAttr1->getSurface()->fillReplace(x*2,y*2,2,2,ZX_ORANGE,  fg);
        m_bmAttr1->getSurface()->fillReplace(x*2,y*2,2,2,ZX_BRORANGE,bg);

        if (mask & 128) { mask = fg; fg = bg; bg = mask; }

        m_bmAttr2->getSurface()->fillReplace(x*2,y*2,2,2,ZX_ORANGE,  fg);
        m_bmAttr2->getSurface()->fillReplace(x*2,y*2,2,2,ZX_BRORANGE,bg);

	drawPixel(x, y);
	drawZoom();
}


int ScreenEditorPage::changeMode(void)
{
	if (m_mode == MD_BITMAP && m_attrs) 
	{
		m_mode = MD_ATTRIB;
		showMode();
		return ZXE_CONTINUE;
	}
	if (m_mode == MD_ATTRIB && m_format != FMT_XBM && m_format != FMT_ROSCRN)
	{
		m_mode = MD_BITMAP;
		showMode();
		return ZXE_CONTINUE;
	}
	return ZXE_CONTINUE;
}

int ScreenEditorPage::changeGrid(void)
{
	int og = m_grid;

	switch(m_grid)
	{
		case GR_NONE:
			if (m_attrs) m_grid = GR_PIXEL;
			else if (m_format != FMT_XBM && m_format != FMT_ROSCRN)
				     m_grid = GR_ATTR;
			break;
		case GR_PIXEL:
			if (m_format != FMT_XBM && m_format != FMT_ROSCRN)
				m_grid = GR_ATTR;
			else	m_grid = GR_NONE;
			break;
		case GR_ATTR:
			m_grid = GR_NONE;
			break;
	}
	if (m_grid != og)
	{
		setup(false);
		drawData(0);		
	}
	return ZXE_CONTINUE;
}


int ScreenEditorPage::onUndo(void)
{
	int n;
	jswByte s;

	if (m_format != FMT_XBM && m_format != FMT_ROSCRN) for (n = (m_w * m_h / 8) - 1; n >= 0; n--)
	{
		s = m_bits[n];
		m_bits[n] = m_undoBits[n];
		m_undoBits[n] = s;
	}
	if (m_attrs) for (n = m_aw * m_ah - 1; n >= 0; n--)
	{
		s = m_attrs[n];
		m_attrs[n] = m_undoAttrs[n];
		m_undoAttrs[n] = s;
	}
	setup(false);
	drawData(0);
	return ZXE_CONTINUE;
}


