/************************************************************************

    JSWED 2.0.1 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "executor.hxx"
#ifdef _WIN32
#include <windows.h>
#endif

#define DOTFILE_DIR ".jswed"

// 2.0.1: BeOS support
#ifndef O_NOFOLLOW
#define O_NOFOLLOW 0
#endif

char *
getResourceName(char *name, char *mode)
{
        FILE *fp;
        static char fname[PATH_MAX];
        char *home;

// 1. Look in the library dir.
        sprintf(fname, "%s/%s", jswLibDir(), name);
	fp = fopen(fname, mode);
        if (fp) { fclose(fp); return fname; }
// 2. Look in the user's home dir.

        home = getenv("HOME");
        if (!home) home = ".";
        sprintf(fname, "%s/%s", home, DOTFILE_DIR);
#ifdef _WIN32
	mkdir(fname);
#else
        mkdir(fname, 0777);
#endif
        sprintf(fname, "%s/%s/%s", home, DOTFILE_DIR, name);
        fp = fopen(fname, mode);
        if (fp) { fclose(fp); return fname; }

// 3. Look in the current directory
        return name;
}

// Wrapper for getResourceName()

FILE *
openResource(char *name, char *mode)
{
        char *s = getResourceName(name, mode);

        return fopen(s, mode);
}


char *
jswLibDir(void)
{
	char *s = getenv("JSWED_LIB");
	
	if (s) return s;
#ifdef _WIN32
        static char buf[PATH_MAX];
 
        GetModuleFileName(NULL, buf, PATH_MAX);
	s = strrchr(buf, '\\');
	if (!s) s = strrchr(buf, '/');
	if (s)
	{
		strcpy(s+1, "Lib");
		return buf;
	}
 
#endif  // _WIN32                                                                       
#ifdef DEBUGGING
	return "Lib";
#else
	return JSW_LIB_DIR;
#endif	// DEBUGGING
}

#ifdef DEBUGGING
void
jswLog (char *s, ...)
{
        //arglist ap;
       	va_list ap; 
        va_start(ap, s);
#ifdef _WIN32
	{
		char b[1000];
		vsprintf(b, s, (va_list)ap);
		MessageBox(NULL, b, "JSWED", MB_OK);
	}
#else
        vprintf(s, (va_list)ap);
#endif
        va_end(ap);
}
#endif

void
jswDiewith(int e, char *s, ...)
{
	//arglist ap;
	va_list ap;
	
	va_start(ap, s);
	vfprintf(stderr, s, (va_list)ap);
	va_end(ap);
	exit(e);
}


FILE *openTemp(char *name, char *type, char *mode)
{
	FILE *fp = NULL;

#ifdef _WIN32
	do
	{
		char tname[PATH_MAX];
		GetTempPath(PATH_MAX, tname);
		if (!GetTempFileName(tname, "z80", 0, name)) return NULL;
		remove(name);

		char *s = strrchr(name, '.');
		if (s) strcpy(s, type);
		else   strcat(name, type);
                // The O_EXCL prevents the possibility of overwriting
                // existing files.
                int fd = open(name, O_RDWR | O_CREAT | O_EXCL | O_BINARY, 
					S_IRUSR | S_IWUSR);
		fp = NULL;
 		if (fd >= 0) fp = fdopen(fd, mode); //
	} while (!fp); 
#else 
	char *tdir = getenv("TMPDIR");
	do
	{
		// I'm using tmpnam() here. By itself this is not 
		// sufficient to guarantee a unique temporary filename;
		// but mkstemp() doesn't allow me to specify a ".z80" 
		// ending, which I need.
		if (tdir) sprintf(name, "%s/jswedXXXXXXXX", tdir);
		else	  sprintf(name, TMPDIR "/jswedXXXXXXXX");

		if (!tmpnam(name)) continue;
		strcat(name, type);

		// The O_EXCL | O_NOFOLLOW prevents symlink attacks on the 
		// possibly insecure temp filename I got. 
		int fd = open(name, O_RDWR | O_CREAT | O_EXCL | O_NOFOLLOW, 
				S_IRUSR | S_IWUSR);
		if (fd < 0) continue;

		fp = fdopen(fd, mode);
		if (!fp) perror(name);
		return fp;
	} while (!fp);
#endif
	return fp;
}



//
// Execute a .Z80 file
//
int executeZ80(char *filename, bool removeAfterwards)
{
	Executor ex(filename, removeAfterwards);	

	return ex.exec();
}


int alert(char *title, char *option)
{
	TickListener *tl = setTickListener(NULL);
	VideoMenu v(title, option, NULL);
	int r = v.doModal();
	setTickListener(tl);
	if (r >= ZXE_QUIT) return r;
	return ZXE_CONTINUE;
}


#ifndef _WIN32
char *getHomeDir(void)
{
	static char buf[PATH_MAX];

	struct passwd *pwd = getpwuid(getuid());
	if (pwd) 
	{
		strcpy(buf, pwd->pw_dir);
		return buf;
	}
	char *hdir = getenv("HOME");
	if (hdir)
	{
		strcpy(buf, hdir);
		return buf;	
	}
	return "/";	// What else can you do?
}
#endif


char *getSpectrumEmulator(void)
{
	static char buf[PATH_MAX + 20];
#ifdef _WIN32
	HKEY hKey;
	DWORD disp;

	long r = RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\jce@seasip\\JSWED2",
			0, "REG_SZ", REG_OPTION_NON_VOLATILE, 
			KEY_ALL_ACCESS, NULL, &hKey, &disp);
	if (r != ERROR_SUCCESS)	return "";
	disp = PATH_MAX;
	r = RegQueryValueEx(hKey, "SpectrumEmulator", NULL, NULL, 
				(BYTE *)buf, &disp);
	if (r != ERROR_SUCCESS) 
	{
		RegCloseKey(hKey);
		return "";
	}
	RegCloseKey(hKey);
	buf[disp] = 0;
#else
	FILE *fp;
	char *homedir = getHomeDir();

	sprintf(buf, "%s/.jswedrc", homedir);
	fp = fopen(buf, "r");
	if (!fp) { buf[0] = 0; return buf; }	// 2.0.1: Don't return path to .jswedrc!
	while (fgets(buf, sizeof(buf), fp))
	{
		if (!strncmp(buf, "SpectrumEmulator", 16))
		{
			char *t = strrchr(buf, '\n');
			if (t) *t = 0;
			char *s = strchr(buf, '=');
			if (!s) continue;
			return s + 1;
		}
	}
	buf[0] = 0;
#endif
	return buf;
}


void setSpectrumEmulator(char *program)
{
#ifdef _WIN32
	HKEY hKey;
	DWORD disp;

	long r = RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\jce@seasip\\JSWED2",
			0, "REG_SZ", REG_OPTION_NON_VOLATILE, 
			KEY_ALL_ACCESS, NULL, &hKey, &disp);
	if (r != ERROR_SUCCESS)	return;
	RegSetValueEx(hKey, "SpectrumEmulator", 0, REG_SZ, (BYTE *)program,
			1 + strlen(program));
	RegCloseKey(hKey);
#else
        char buf[PATH_MAX];
	FILE *fp;
	char *homedir = getHomeDir();

	sprintf(buf, "%s/.jswedrc", homedir);
	fp = fopen(buf, "w");
	if (!fp) return;
	fprintf(fp, "SpectrumEmulator=%s\n", program);
	fclose(fp);
#endif
}

